# coding=utf-8
# Copyright 2019 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import with_statement
from __future__ import absolute_import
import pytest

import numpy as np

import cirq
from itertools import izip


class NoMethod(object):
    pass


class ReturnsNotImplemented(object):
    def _mixture_(self):
        return NotImplemented

    def _has_mixture_(self):
        return NotImplemented


class ReturnsValidTuple(cirq.SupportsMixture):
    def _mixture_(self):
        return ((0.4, u'a'), (0.6, u'b'))

    def _has_mixture_(self):
        return True


class ReturnsNonnormalizedTuple(object):
    def _mixture_(self):
        return ((0.4, u'a'), (0.4, u'b'))


class ReturnsNegativeProbability(object):
    def _mixture_(self):
        return ((0.4, u'a'), (-0.4, u'b'))


class ReturnsGreaterThanUnityProbability(object):
    def _mixture_(self):
        return ((1.2, u'a'), (0.4, u'b'))


class ReturnsMixtureButNoHasMixture(object):
    def _mixture_(self):
        return ((0.4, u'a'), (0.6, u'b'))


class ReturnsUnitary(object):
    def _unitary_(self):
        return np.ones((2, 2))

    def _has_unitary_(self):
        return True


class ReturnsNotImplementedUnitary(object):
    def _unitary_(self):
        return NotImplemented

    def _has_unitary_(self):
        return NotImplemented

@pytest.mark.parametrize(u'val', (NoMethod(), ReturnsNotImplemented(),))
def test_objects_with_no_mixture(val):
    with pytest.raises(TypeError, match=u"mixture"):
        _ = cirq.mixture(val)
    assert cirq.mixture(val, None) is None
    assert cirq.mixture(val, NotImplemented) is NotImplemented
    default = ((0.4, u'a'), (0.6, u'b'))
    assert cirq.mixture(val, default) == default


@pytest.mark.parametrize(u'val,mixture', (
    (ReturnsValidTuple(), ((0.4, u'a'), (0.6, u'b'))),
    (ReturnsNonnormalizedTuple(), ((0.4, u'a'), (0.4, u'b'))),
))
def test_objects_with_mixture(val, mixture):
    assert cirq.mixture(val) == mixture
    assert cirq.mixture(val, ((0.3, u'a'), (0.7, u'b'))) == mixture


@pytest.mark.parametrize(u'val,mixture', (
    (ReturnsValidTuple(), ((0.4, u'a'), (0.6, u'b'))),
    (ReturnsNonnormalizedTuple(), ((0.4, u'a'), (0.4, u'b'))),
    (ReturnsUnitary(), ((1.0, np.ones((2, 2))),))
))
def test_objects_with_mixture_channel(val, mixture):
    expected_keys, expected_values = izip(*mixture)
    keys, values = izip(*cirq.mixture_channel(val))
    np.testing.assert_almost_equal(keys, expected_keys)
    np.testing.assert_equal(values, expected_values)

    keys, values = izip(*cirq.mixture_channel(val, ((0.3, u'a'), (0.7, u'b'))))
    np.testing.assert_almost_equal(keys, expected_keys)
    np.testing.assert_equal(values, expected_values)


@pytest.mark.parametrize(u'val', (
NoMethod(), ReturnsNotImplemented(), ReturnsNotImplementedUnitary()))
def test_objects_with_no_mixture_channel(val):
    with pytest.raises(TypeError, match=u"mixture"):
        _ = cirq.mixture_channel(val)
    assert cirq.mixture(val, None) is None
    assert cirq.mixture(val, NotImplemented) is NotImplemented
    default = ((0.4, u'a'), (0.6, u'b'))
    assert cirq.mixture(val, default) == default


def test_has_mixture():
    assert cirq.has_mixture(ReturnsValidTuple())
    assert not cirq.has_mixture(ReturnsNotImplemented())
    assert cirq.has_mixture(ReturnsMixtureButNoHasMixture())
    assert not cirq.has_mixture(ReturnsUnitary())


def test_has_mixture_channel():
    assert cirq.has_mixture_channel(ReturnsValidTuple())
    assert not cirq.has_mixture_channel(ReturnsNotImplemented())
    assert cirq.has_mixture_channel(ReturnsMixtureButNoHasMixture())
    assert cirq.has_mixture_channel(ReturnsUnitary())
    assert not cirq.has_mixture_channel(ReturnsNotImplementedUnitary())


def test_valid_mixture():
    cirq.validate_mixture(ReturnsValidTuple())


@pytest.mark.parametrize(u'val,message', (
    (ReturnsNonnormalizedTuple(), u'1.0'),
    (ReturnsNegativeProbability(), u'less than 0'),
    (ReturnsGreaterThanUnityProbability(), u'greater than 1')
))
def test_invalid_mixture(val, message):
    with pytest.raises(ValueError, match=message):
        cirq.validate_mixture(val)


def test_missing_mixture():
    with pytest.raises(TypeError, match=u'_mixture_'):
        cirq.validate_mixture(NoMethod)
